"""The pull function for fetching the economic data."""

# pylint: disable=line-too-long
import datetime
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any

import pandas as pd
import requests_cache
import tqdm
from requests_cache.session import CachedSession

from .coinbase import pull as coinbase_pull
from .data import Data
from .efd import pull as efd_pull
from .fred import pull as fred_pull
from .oecd import pull as oecd_pull
from .yfinance import pull as yfinance_pull

_DEFAULT_MANIFEST = {
    str(Data.FRED): {
        "ECBESTRVOLWGTTRMDMNRT": True,  # Euro Short-Term Rate: Volume-Weighted Trimmed Mean Rate
        "SOFR30DAYAVG": True,  # 30-Day Average SOFR
        "SOFR": True,  # Secured Overnight Financing Rate"
        "ADPWNUSNERSA": True,  # Total Nonfarm Private Payroll Employment
        "GDP": True,  # Gross Domestic Product
        "ECIWAG": True,  # Employment Cost Index: Wages and Salaries: Private Industry Workers
        "NFCI": True,  # Chicago Fed National Financial Conditions Index
        "EFFR": True,  # Effective Federal Funds Rate
        "DTP10J28": True,  # 10-Year 0.5% Treasury Inflation-Indexed Note
        "AMERIBOR": True,  # Overnight Unsecured AMERIBOR Benchmark Interest Rate
        "SOFR90DAYAVG": True,  # 90-Day Average SOFR
        "SOFRVOL": True,  # Secured Overnight Financing Volume
        "DFEDTARU": True,  # Federal Funds Target Range - Upper Limit
        "IUDSOIA": True,  # Daily Sterling Overnight Index Average (SONIA) Rate
        "BBKMGDP": True,  # Brave-Butters-Kelley Real Gross Domestic Product
        "ECBDFR": True,  # ECB Deposit Facility Rate for Euro Area
        "ECBESTRTOTVOL": True,  # Euro Short-Term Rate: Total Volume
        "DFEDTARL": True,  # Federal Funds Target Range - Lower Limit
        "CORESTICKM159SFRBATL": True,  # Sticky Price Consumer Price Index less Food and Energy
        "CPIAUCSL": True,  # Consumer Price Index for All Urban Consumers: All Items in U.S. City Average
        "GDPC1": True,  # Real Gross Domestic Product
        "MEDCPIM158SFRBCLE": True,  # Median Consumer Price Index
        "USACPALTT01CTGYM": True,  # Consumer Price Indices (CPIs, HICPs), COICOP 1999: Consumer Price Index: Total for United States
        "CPALTT01USM657N": True,  # Consumer Price Index: All Items: Total for United States
    },
    str(Data.YFINANCE): {
        # Equities
        "MSFT": True,  # Microsoft
        "NVDA": True,  # NVIDIA
        "F": True,  # Ford
        "AAPL": True,  # Apple
        "LCID": True,  # Lucid Group, Inc
        "AMZN": True,  # Amazon
        "PLTR": True,  # Palantir
        "APLD": True,  # Applied Digital Corporation
        "GOOG": True,  # Alphabet
        "TSLA": True,  # Tesla
        "META": True,  # Meta
        "WBD": True,  # Warner Bros. Discovery, Inc.
        "2222.SR": True,  # Saudi Arabian Oil Company
        "INTC": True,  # Intel Corporation
        "AVGO": True,  # Broadcom
        "GRAB": True,  # Grab Holdings Limited
        "TSM": True,  # Taiwan Semiconductor Manufacturing Company
        # Forex
        "EURUSD=X": True,  # Euro to USD
        "JPY=X": True,  # Yen to USD
        "GBPUSD=X": True,  # Sterling to USD
        "AUDUSD=X": True,  # AUD to USD
        "NZDUSD=X": True,  # NZD to USD
        "EURJPY=X": True,  # Euro to Yen
        "GBPJPY=X": True,  # Sterling to Yen
        # Futures
        "ES=F": True,  # E-Mini S&P 500
        "YM=F": True,  # Mini Dow Jones
        "NQ=F": True,  # Nasdaq 100
        "RTY=F": True,  # E-mini Russell 2000
        "ZB=F": True,  # U.S. Treasury Bond Futures
        "ZN=F": True,  # 10-Year T-Note Futures
    },
    str(Data.COINBASE): {
        "BTC-USD": True,  # Bitcoin
        "XRP-USD": True,  # Ripple
        "ETH-USD": True,  # Ethereum
        "USDT-USD": True,  # Tether
        "SOL-USD": True,  # Solana
        # "USDC-USD": True,  # USDC
        "DOGE-USD": True,  # DOGE
        # "TRX-USD": True,  # TRON
        "ADA-USD": True,  # Cardano
    },
    str(Data.EFD): {
        "AAPL": True,  # Apple Senate Trades
        "MSFT": True,  # Microsoft Senate Trades
        "NVDA": True,  # Nvidia Senate Trades
        "F": True,  # Ford Senate Trades
        "LCID": True,  # Lucid Group, Inc
        "AMZN": True,  # Amazon
        "PLTR": True,  # Palantir
        "APLD": True,  # Applied Digital Corporation
    },
    str(Data.OECD): {
        "OECD.TAD.ATM,DSD_AGR@DF_OUTLOOK_2021_2030,1.0/NOR.A.CPC_0141...": True,  # Agricultural Outlook Norway Soy Beans
        "OECD.TAD.ATM,DSD_AGR@DF_OUTLOOK_2021_2030,1.0/AUS.A.CPC_0141...": True,  # Agricultural Outlook Australia Soy Beans
        "OECD.TAD.ATM,DSD_AGR@DF_OUTLOOK_2021_2030,1.0/CAN.A.CPC_0141...": True,  # Agricultural Outlook Canada Soy Beans
    },
}
_DATA_PROVIDERS = {
    str(Data.FRED): fred_pull,
    str(Data.YFINANCE): yfinance_pull,
    str(Data.COINBASE): coinbase_pull,
    str(Data.EFD): efd_pull,
    str(Data.OECD): oecd_pull,
}


def pull(
    manifest: dict[str, dict[str, Any]] | None = None,
    min_date: datetime.date | None = None,
    session: CachedSession | None = None,
) -> pd.DataFrame:
    """Pull the latest economic data."""
    if manifest is None:
        manifest = _DEFAULT_MANIFEST
    if session is None:
        session = CachedSession(
            "econball",
            expire_after=requests_cache.NEVER_EXPIRE,
            allowable_methods=("GET", "HEAD", "POST"),
            stale_if_error=True,
        )
    series_pool = []
    with ThreadPoolExecutor() as executor:
        futures = []
        for k, v in _DEFAULT_MANIFEST.items():
            futures.extend([executor.submit(_DATA_PROVIDERS[k], x, session) for x in v])
        for future in tqdm.tqdm(as_completed(futures), desc="Downloading"):
            series = future.result()
            series_pool.extend(series)
    series_pool = [
        s
        if isinstance(s.index, pd.DatetimeIndex)
        else pd.Series(s.values, index=pd.to_datetime(s.index), name=s.name)
        for s in series_pool
    ]
    df = pd.concat(series_pool, axis=1).sort_index().asfreq("D", method="ffill").ffill()
    if min_date is not None:
        df = df[df.index.date >= min_date]  # type: ignore
    df = df[df.index.date < datetime.datetime.today().date()]  # type: ignore
    return df



"""A puller for coinbase data."""

# pylint: disable=unused-argument
import datetime
import logging

import pandas as pd
from cbhist.historical import fetch_historical  # type: ignore
from requests_cache.session import CachedSession


def pull(series: str, session: CachedSession) -> list[pd.Series]:
    """Pull the coinbase data."""
    logging.info("Pulling coinbase series %s", series)
    df = fetch_historical(series, 60 * 60 * 24, datetime.datetime(2000, 1, 1))
    df.index = df.index.date  # type: ignore
    df = df.sort_index()
    return [
        df["close"].rename(series + "_close"),
        df["volume"].rename(series + "_volume"),
    ]



"""A puller for FRED data."""

# pylint: disable=global-statement,unused-argument
import logging
import os

import pandas as pd
from fredapi import Fred  # type: ignore
from requests_cache.session import CachedSession

_FRED_CLIENT = None


def _get_fred_client() -> Fred:
    global _FRED_CLIENT
    if _FRED_CLIENT is None:
        _FRED_CLIENT = Fred(api_key=os.environ["FRED_API_KEY"])
    return _FRED_CLIENT


def pull(series: str, session: CachedSession) -> list[pd.Series]:
    """Pull the FRED economic data."""
    logging.info("Pulling FRED series %s", series)
    series_name = "FRED_" + series
    client = _get_fred_client()
    try:
        df = client.get_series_all_releases(series)
    except ValueError:
        df = client.get_series(series)
        df.index = df.index.date  # type: ignore
        df = df.sort_index()
        return [df.rename(series_name)]
    df["date"] = pd.to_datetime(df["date"])
    df["realtime_start"] = pd.to_datetime(df["realtime_start"])

    def select_latest(group: pd.DataFrame) -> pd.DataFrame:
        return group[group["realtime_start"] == group["realtime_start"].max()]

    df = df.groupby("date").apply(select_latest)
    df = df.set_index("date")
    df.index = df.index.date  # type: ignore
    df = df.sort_index()
    return [df["value"].rename(series_name)]



"""An enumeration for the different data sources in econball."""

from enum import StrEnum, auto


class Data(StrEnum):
    """Data enumeration."""

    FRED = auto()
    YFINANCE = auto()
    COINBASE = auto()
    EFD = auto()
    OECD = auto()



"""The CLI for executing the economics data pull."""

import io
import logging
import sys

from dotenv import load_dotenv

from . import __VERSION__
from .args import parse_args
from .pull import pull

_STDOUT_FILE = "-"


def main() -> None:
    """The main CLI function."""
    load_dotenv()
    logging.basicConfig(level=logging.INFO)
    args = parse_args()

    logging.info("--- econball %s ---", __VERSION__)

    df = pull(min_date=args.min_date)
    logging.info(df)

    handle = io.BytesIO()
    df.to_parquet(handle, compression="gzip")
    handle.seek(0)
    if args.file == _STDOUT_FILE:
        sys.stdout.buffer.write(handle.getbuffer())
    else:
        with open(args.file, "wb") as fhandle:
            fhandle.write(handle.getbuffer())


if __name__ == "__main__":
    main()



"""A puller for EFD data."""

# pylint: disable=global-statement,line-too-long,too-many-locals,too-many-branches,too-many-statements
import io
import logging
import multiprocessing
import re
import urllib.parse
from urllib.parse import urlparse

import pandas as pd
from bs4 import BeautifulSoup
from dateutil import parser
from requests_cache.session import CachedSession

_EFD_DATA = None
_LOCK = multiprocessing.Lock()


def _fetch_efd_date(
    session: CachedSession,
) -> dict[str, dict[str, list[dict[str, float | str]]]]:
    with _LOCK:
        # Adapted from https://github.com/neelsomani/senator-filings/tree/master
        global _EFD_DATA
        if _EFD_DATA is None:
            urls = set()
            with session.cache_disabled():
                landing_page_response = session.get(
                    "https://efdsearch.senate.gov/search/home/"
                )
                landing_page_response.raise_for_status()
                landing_page = BeautifulSoup(landing_page_response.text, "lxml")
                form_csrf = landing_page.find(attrs={"name": "csrfmiddlewaretoken"})[  # type: ignore
                    "value"
                ]  # type: ignore
                form_payload = {
                    "csrfmiddlewaretoken": form_csrf,
                    "prohibition_agreement": "1",
                }
                session.post(
                    "https://efdsearch.senate.gov/search/home/",
                    data=form_payload,
                    headers={"Referer": "https://efdsearch.senate.gov/search/home/"},
                )
                if "csrftoken" in session.cookies:
                    csrftoken = session.cookies["csrftoken"]
                else:
                    csrftoken = session.cookies["csrf"]

                def reports_api(
                    offset: int,
                ) -> list[list[str]]:
                    login_data = {
                        "start": str(offset),
                        "length": str(100),
                        "report_types": "[11]",
                        "filer_types": "[]",
                        "submitted_start_date": "01/01/2012 00:00:00",
                        "submitted_end_date": "",
                        "candidate_state": "",
                        "senator_state": "",
                        "office_id": "",
                        "first_name": "",
                        "last_name": "",
                        "csrfmiddlewaretoken": csrftoken,
                    }
                    response = session.post(
                        "https://efdsearch.senate.gov/search/report/data/",
                        data=login_data,
                        headers={"Referer": "https://efdsearch.senate.gov/search/"},
                    )
                    return response.json()["data"]

                idx = 0
                reports = reports_api(idx)
                all_reports: list[list[str]] = []
                while len(reports) != 0:
                    all_reports.extend(reports)
                    idx += 100
                    reports = reports_api(idx)

            for report in all_reports:
                for line in report:
                    match = re.search(
                        r'<a\s[^>]*href=["\'](.*?)["\']', line, re.IGNORECASE
                    )
                    if match:
                        href = match.group(1)
                        url = urllib.parse.urljoin(
                            "https://efdsearch.senate.gov/search/report/data/", href
                        )
                        urls.add(url)

            traded_assets: dict[str, dict[str, list[dict[str, float | str]]]] = {}

            for url in urls:
                o = urlparse(url)
                path_components = o.path.split("/")
                if path_components[-3] != "ptr":
                    continue

                response = session.get(url)
                response.raise_for_status()

                handle = io.StringIO()
                handle.write(response.text)
                handle.seek(0)
                dfs = pd.read_html(handle)
                df = dfs[0]

                soup = BeautifulSoup(response.text, "lxml")
                filed_date = None
                for h1 in soup.find_all("h1"):
                    filed_date = parser.parse(
                        h1.get_text().strip().split("(")[0].strip().split()[-1]
                    )
                if filed_date is None:
                    raise ValueError("filed_date is null")

                for _, row in df.iterrows():
                    ticker = row["Ticker"]
                    if "--" in ticker:
                        continue
                    date = str(filed_date)
                    amount = float(
                        row["Amount"]
                        .split("-")[0]
                        .strip()
                        .replace(",", "")
                        .replace("$", "")
                    )
                    interaction_type = row["Type"].split("(")[0].strip()
                    trade_interaction_type = None
                    if interaction_type in {"Purchase", "Exchange"}:
                        trade_interaction_type = "buy"
                    elif interaction_type == "Sale":
                        trade_interaction_type = "sale"
                    else:
                        raise ValueError(
                            f"Unrecognised interaction type: {interaction_type}"
                        )
                    timeline = traded_assets.get(ticker, {})
                    trades = timeline.get(date, [])
                    trades.append(
                        {"amount": amount, "interaction": trade_interaction_type}
                    )
                    timeline[date] = trades
                    traded_assets[ticker] = timeline

            _EFD_DATA = traded_assets
        return _EFD_DATA


def pull(series: str, session: CachedSession) -> list[pd.Series]:
    """Pull the EFD data."""
    logging.info("Pulling EFD series %s", series)
    data = _fetch_efd_date(session)
    ticker = data.get(series)
    if ticker is None:
        return []
    actions = set()
    for value in ticker.values():
        for trade in value:
            actions.add(trade["interaction"])
    action_series = []
    for action in actions:
        summed_dates = {}
        for k, v in ticker.items():
            summed_amounts = 0.0
            for trade in v:
                if trade["interaction"] != action:
                    continue
                summed_amounts += trade["amount"]  # type: ignore
            if summed_amounts == 0.0:
                continue
            summed_dates[k] = summed_amounts
        action_ser = pd.Series(
            data=list(summed_dates.values()), index=list(summed_dates.keys())
        )
        action_ser.index = pd.to_datetime(action_ser.index).date  # type: ignore
        action_ser.name = f"EFD_{series}_{action}"
        action_ser = action_ser.sort_index()
        action_series.append(action_ser)
    return action_series



"""A puller for OECD data using pandasdmx."""

# pylint: disable=unused-argument,line-too-long,unnecessary-lambda,too-many-locals
import logging
from typing import List

import pandas as pd
import requests
from requests_cache.session import CachedSession


def pull(series: str, session: CachedSession) -> List[pd.Series]:
    """Pull time series from the OECD."""
    logging.info("Pulling OECD series %s", series)

    url = f"https://sdmx.oecd.org/public/rest/data/{series}?startPeriod=2010&endPeriod=2030&format=jsondata"
    response = requests.get(url, timeout=30.0)
    response.raise_for_status()
    data = response.json()
    data_val = data["data"]

    dimensions = data_val["structures"][0]["dimensions"]
    names = [
        x["name"]
        for x in [x["values"] for x in dimensions["series"] if x["id"] == "MEASURE"][0]
    ]
    ts_index = pd.DatetimeIndex(
        pd.to_datetime(
            [
                x["start"]
                for x in data_val["structures"][0]["dimensions"]["observation"][0][
                    "values"
                ]
            ]
        )
    )

    dataset_series = data_val["dataSets"][0]["series"]
    series_list = []
    for count, dataset_key in enumerate(dataset_series):
        observations = dataset_series[dataset_key]["observations"]
        observation_keys = sorted(observations.keys(), key=lambda x: int(x))

        try:
            series_list.append(
                pd.Series(
                    name="_".join([series, names[count]]),
                    data=[float(observations[x][0]) for x in observation_keys],
                    index=ts_index,
                )
            )
        except ValueError as exc:
            logging.warning(str(exc))
            logging.warning("Failed to process %s", series)

    return series_list



"""Find the args used by the main CLI."""

import argparse
import datetime

STDOUT_FILE = "-"


def _valid_date(s: str) -> datetime.date:
    try:
        return datetime.datetime.strptime(s, "%Y-%m-%d").date()
    except ValueError as exc:
        raise argparse.ArgumentTypeError(f"not a valid date: {s!r}") from exc


def parse_args() -> argparse.Namespace:
    """Create the args based on the CLI inputs."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "file",
        default=STDOUT_FILE,
        help="The file to write the output to (- if to stdout).",
    )
    parser.add_argument(
        "--min-date",
        required=False,
        help="The minimum date (YYYY-MM-DD) to consider when generating data.",
        type=_valid_date,
    )
    return parser.parse_args()



"""A puller for yahoo finance data."""

# pylint: disable=global-statement,unused-argument
import logging

import curl_cffi
import pandas as pd
import yfinance as yf  # type: ignore
from requests_cache.session import CachedSession


def pull(series: str, session: CachedSession) -> list[pd.Series]:
    """Pull the yfinance data."""
    logging.info("Pulling yfinance series %s", series)
    series_name = "yfinance_" + series
    try:
        df = yf.Ticker(series).history(start="2000-01-01")
        df.index = df.index.date  # type: ignore
        df = df.sort_index()
        return [
            df["Close"].rename(series_name + "_close"),
            df["Volume"].rename(series_name + "_volume"),
        ]
    except curl_cffi.requests.exceptions.HTTPError as exc:
        logging.warning(str(exc))
        logging.warning("Could not pull %s", series)
        return []
